#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- Python -*-

"""
 @file DroneIO.py
 @brief Wrapper of drone flight controller(ArduPilot)
 @date $Date$

 @author East Japan Accounting Center Co., Ltd

 BSD 3-clause "New" or "Revised" License

"""
import sys
import time
import socket
import math

# Import RTM module
import RTC
import OpenRTM_aist

# drone distance calc
import drone_distance_cal

# Setup logger
from logging import getLogger, StreamHandler
from logging import DEBUG as LOG_LV

# Import DroneKit
from dronekit import connect, VehicleMode, LocationGlobalRelative, Command, Vehicle
from pymavlink import mavutil

sys.path.append(".")

logger = getLogger(__name__)
handler = StreamHandler()
logger.setLevel(LOG_LV)
handler.setLevel(LOG_LV)
logger.addHandler(handler)

# Constant declaration
TARGET_IP = "127.0.0.1:14550"  # Vehicle target address
MIN_MV_DIST = 2.0  # minimum move distance
HOME_LOC_GET_RETRY_CNT = 3  # The max num of downloads of the "home_location".
ARMED_RETRY_CNT = 2  # the max wait time of reflect "armed"
COMM_ERR_TIMER = 30  # communication error(w/ the drone controller
TAKEOFF_ARV_COEF = 0.95  # Takeoff arrival coefficient
TGT_ALT_ARV_COEF_MIN = 0.95  # Target alt arrival coefficient minimum
TGT_ALT_ARV_COEF_MAX = 1.05  # Target alt arrival coefficient maximum
VEHICLE_DEF_AIR_SPD = 3  # vehicle default speed(unit:m/s)
BATT_LIPO_VPC = 3.7  # Lipo Battery VPC(voltage per cell)
# Lipo Battery VPC(voltage per cell) percentage It means 3.7V = 50%
BATT_LIPO_VPC_PERCENT = 0.5
# Battery Parameters Tuning
BATT_MONI_V_ONLY = 3  # Analog Voltage Only
BATT_V_PIN_PIXHAWK = 2  # Battery Voltage sensing pin(Pixhawk)
BATT_C_PIN_PIXHAWK = -1  # Battery Current sensing pin(Disabled)
# RTL cause
RTL_CAUSE_BY_MANU = 1
RTL_CAUSE_BY_ERR = 2
# manual RTL stage
RTL_STAGE_1ST = 1  # Goto RTL altitude
RTL_STAGE_2ND = 2  # Goto home-pos and keeping altitude
RTL_STAGE_3RD = 3  # Landing on the home-pos

# Error message
ERR_MSG_001 = "[E]Socket error. Please re-startup this RTC."
ERR_MSG_002 = "[E]Connection Timeout. Please re-startup this RTC."
ERR_MSG_003 = "[E]home_location timeout. Please re-startup this RTC."
ERR_MSG_004 = "[E]Does not the home_location. Please re-startup this RTC."
ERR_MSG_005 = "[E]Arm switch timeout. Please re-startup this RTC."
ERR_MSG_006 = "[E]Communication with the GCS has been lost. Start the RTL."
ERR_MSG_007 = "[E]Low battery detected. Start the RTL."

# drone status declaration
STS_DISARM = 0
STS_ARMED = 1
STS_TAKEOFF = 2
STS_RTL = 3
STS_IN_MOVE = 4
STS_ALT_CHG = 5
STS_HOVERING = 6
STS_YAW_CHG = 7

# Copter parameter
#   Flight Mode
COPTER_PARAM_FLTMODE_STABILIZE = 0.0
COPTER_PARAM_FLTMODE_GUIDED = 4.0
COPTER_PARAM_FLTMODE_LOITER = 5.0

# drone status declaration(string)
droneio_sts_str = ("Disarm",
                   "Armed",
                   "Takeoff",
                   "RTL",
                   "In_Move",
                   "Alt_CHG",
                   "Hovering",
                   "Yaw_CHG"
                   )

# Import Service implementation class
# <rtc-template block="service_impl">

# </rtc-template>

# Import Service stub modules
# <rtc-template block="consumer_import">
# </rtc-template>


# This module's spesification
# <rtc-template block="module_spec">
droneio_spec = ["implementation_id", "DroneIO", 
         "type_name",         "DroneIO", 
         "description",       "Wrapper of drone flight controller(ArduPilot)", 
         "version",           "1.3.0", 
         "vendor",            "EAC", 
         "category",          "Drone", 
         "activity_type",     "STATIC", 
         "max_instance",      "1", 
         "language",          "Python", 
         "lang_type",         "SCRIPT",
         "conf.default.BattCell", "4",
         "conf.default.RTLLowBatt", "30",
         "conf.default.BattChk", "on",
         "conf.default.TakeoffAlt", "10.0",
         "conf.default.RTLAlt", "1500",
         "conf.default.RTLAltFinal", "0",
         "conf.default.WpNavSpeed", "500",
         "conf.default.RTLLoitTime", "5000",
         "conf.default.WpNavSpeedDn", "100",
         "conf.default.LandSpeed", "50",
         "conf.default.DisarmDelay", "30",
         "conf.default.SystemBehavior", "single",
         "conf.__widget__.BattCell", "spin.1",
         "conf.__widget__.RTLLowBatt", "spin.1",
         "conf.__widget__.BattChk", "radio",
         "conf.__widget__.TakeoffAlt", "spin.0.1",
         "conf.__widget__.RTLAlt", "spin.1",
         "conf.__widget__.RTLAltFinal", "spin.1",
         "conf.__widget__.WpNavSpeed", "spin.50",
         "conf.__widget__.RTLLoitTime", "spin.5000",
         "conf.__widget__.WpNavSpeedDn", "spin.10",
         "conf.__widget__.LandSpeed", "spin.10",
         "conf.__widget__.DisarmDelay", "spin.1",
         "conf.__widget__.SystemBehavior", "radio",
         "conf.__constraints__.BattCell", "2<=x<=7",
         "conf.__constraints__.RTLLowBatt", "20<=x<=60",
         "conf.__constraints__.BattChk", "(off,on)",
         "conf.__constraints__.TakeoffAlt", "x>=1.0",
         "conf.__constraints__.RTLAlt", "0<=x<=8000",
         "conf.__constraints__.RTLAltFinal", "-1<=x<=1000",
         "conf.__constraints__.WpNavSpeed", "0<=x<=2000",
         "conf.__constraints__.RTLLoitTime", "0<=x<=60000",
         "conf.__constraints__.WpNavSpeedDn", "0<=x<=500",
         "conf.__constraints__.LandSpeed", "30<=x<=200",
         "conf.__constraints__.DisarmDelay", "0<=x<=127",
         "conf.__constraints__.SystemBehavior", "(single,formation)",
         ""]
# </rtc-template>

##
# @class DroneIO
# @brief Wrapper of drone flight controller(ArduPilot)
# 
# Overview:  Refer to the "機能仕様書_DroneIO_RTC.pdf" for details.
# 
# In / Out:  Refer to the "機能仕様書_DroneIO_RTC.pdf" for details.
# 
# Algorythm:  Refer to the "機能仕様書_DroneIO_RTC.pdf" for details.
# 
# 
class DroneIO(OpenRTM_aist.DataFlowComponentBase):
    
    ##
    # @brief constructor
    # @param manager Maneger Object
    # 
    def __init__(self, manager):
        OpenRTM_aist.DataFlowComponentBase.__init__(self, manager)

        self._d_inTargetPos = RTC.GPSData(
            RTC.Time(0, 0),
            RTC.GPSTime(0, 0),
            0.0, 0.0, 0.0,
            0.0, 0.0,
            0.0, 0.0,
            0.0,
            0,
            RTC.GPS_FIX_NORMAL)
        """
        Drone flight target position
         - Semantics: - Latitude
                      - Longitude
                      - Altitude
                      - air speed(m/s)
        """
        self._inTargetPosIn = OpenRTM_aist.InPort("inTargetPos", self._d_inTargetPos)
        self._d_inTargetSpeed = RTC.TimedFloat(RTC.Time(0,0),0)
        """
        Drone flight target speed
         - Semantics: air speed
         - Unit: m/s
        """
        self._inTargetSpeedIn = OpenRTM_aist.InPort("inTargetSpeed", self._d_inTargetSpeed)
        self._d_inTargetYaw = RTC.TimedShortSeq(RTC.Time(0,0),[])
        """
        Drone flight target yaw
         - Semantics: data[0] direction(1:right, -1:left, 0:none)
                      data[1] relative offset yaw(deg)
        """
        self._inTargetYawIn = OpenRTM_aist.InPort("inTargetYaw", self._d_inTargetYaw)
        self._d_outGPSHome = RTC.GPSData(
            RTC.Time(0, 0),
            RTC.GPSTime(0, 0),
            0.0, 0.0, 0.0,
            0.0, 0.0,
            0.0, 0.0,
            0.0,
            0,
            RTC.GPS_FIX_NORMAL)
        """
        GPS data for "home_location"
         - Semantics: - Latitude
                      - Longitude
                      - Altitude(is relative to mean sea-level)
         - Unit: deg
        """
        self._outGPSHomeOut = OpenRTM_aist.OutPort("outGPSHome", self._d_outGPSHome)
        self._d_outGPSDrone = RTC.GPSData(
            RTC.Time(0, 0),
            RTC.GPSTime(0, 0),
            0.0, 0.0, 0.0,
            0.0, 0.0,
            0.0, 0.0,
            0.0,
            0,
            RTC.GPS_FIX_NORMAL)
        """
        GPS data for "Drone"
         - Semantics: - Latitude
                      - Longitude
                      - Altitude(is relative to the "home_location")
         - Unit: deg
        """
        self._outGPSDroneOut = OpenRTM_aist.OutPort("outGPSDrone", self._d_outGPSDrone)
        self._d_outBattV = RTC.TimedFloat(RTC.Time(0,0),0)
        """
        Battery voltage
         - Unit: V
        """
        self._outBattVOut = OpenRTM_aist.OutPort("outBattV", self._d_outBattV)
        self._d_outBattA = RTC.TimedFloat(RTC.Time(0,0),0)
        """
        Battery current
         - Unit: mA
        """
        self._outBattAOut = OpenRTM_aist.OutPort("outBattA", self._d_outBattA)
        self._d_outFltTime = RTC.Time(RTC.Time(0,0),0)
        """
        Drone flight time
         - Unit: min:sec
        """
        self._outFltTimeOut = OpenRTM_aist.OutPort("outFltTime", self._d_outFltTime)
        self._d_outGyro = RTC.TimedOrientation3D(
            RTC.Time(0, 0),
            RTC.Orientation3D(0.0, 0.0, 0.0))
        """
        3 axis gyro sensor
         - Semantics: - Roll angle
                      - Pitch angle
                      - Yaw angle
         - Unit: rad
        """
        self._outGyroOut = OpenRTM_aist.OutPort("outGyro", self._d_outGyro)
        self._d_outAccel = RTC.TimedFloatSeq(RTC.Time(0,0),[])
        """
        3 axis acceleration sensor
         - Semantics: - x
                      - y
                      - z
         - Unit: m/sec
        """
        self._outAccelOut = OpenRTM_aist.OutPort("outAccel", self._d_outAccel)
        self._d_outMag = RTC.TimedFloat(RTC.Time(0,0),0)
        """
        Magnetic Compass(=Drone's heading)
         - Semantics: 0-360.
                      (North is 0, East is 90, South is 180, West is 270)
         - Unit: deg
        """
        self._outMagOut = OpenRTM_aist.OutPort("outMag", self._d_outMag)
        self._d_outTemp = RTC.TimedFloat(RTC.Time(0,0),0)
        """
        Temperature
         - Unit: degree Celsius
        """
        self._outTempOut = OpenRTM_aist.OutPort("outTemp", self._d_outTemp)
        self._d_outPres = RTC.TimedFloat(RTC.Time(0,0),0)
        """
        Pressure
         - Unit: hPa
        """
        self._outPresOut = OpenRTM_aist.OutPort("outPres", self._d_outPres)
        self._d_outFltMode = RTC.TimedString(RTC.Time(0,0),0)
        """
        Current flight mode
        """
        self._outFltModeOut = OpenRTM_aist.OutPort("outFltMode", self._d_outFltMode)
        self._d_outState = RTC.TimedString(RTC.Time(0,0),0)
        """
        Drone I/O RTC internal mode
         - Semantics: For detail, see the "機能仕様書_DroneIO_RTC.docx"
        """
        self._outStateOut = OpenRTM_aist.OutPort("outState", self._d_outState)
        self._d_outSysErr = RTC.TimedString(RTC.Time(0,0),0)
        """
        Drone I/O RTC error info.
         - Semantics: For detail, see the "機能仕様書_DroneIO_RTC.docx"
        """
        self._outSysErrOut = OpenRTM_aist.OutPort("outSysErr", self._d_outSysErr)
        self._d_outRawIMU = RTC.TimedLongSeq(RTC.Time(0,0),[])
        """
        9axis raw data
        """
        self._outRawIMUOut = OpenRTM_aist.OutPort("outRawIMU", self._d_outRawIMU)
        self._d_outCorrectIMU = RTC.TimedFloatSeq(RTC.Time(0,0),[])
        """
        Based on IMU raw data, will be corrected.
         - Semantics: "outAccel", "outGyro" and "outMag".
         - Unit: accel:m/s gyro:rad mag:deg
        """
        self._outCorrectIMUOut = OpenRTM_aist.OutPort("outCorrectIMU", self._d_outCorrectIMU)

        # initialize of configuration-data.
        # <rtc-template block="init_conf_param">
        """
        The number of cells of the battery mounted on the drone.
         - Name: BattCell BattCell
         - DefaultValue: 4
         - Unit: cell
         - Range: 2<=x<=7
        """
        self._BattCell = [4]
        """
        RTL is done when the battery level falls below the threshold.
         - Name: RTLLowBatt RTLLowBatt
         - DefaultValue: 30
         - Unit: %
         - Range: 20<=x<=60
        """
        self._RTLLowBatt = [30]
        """
        Whether the battery check
         - Name: BattChk BattChk
         - DefaultValue: on
         - Range: (off, on)
        """
        self._BattChk = ['on']
        """
        Takeoff  altitude
         - Name: TakeoffAlt TakeoffAlt
         - DefaultValue: 10.0
         - Unit: m
         - Range: 1.0<=x
        """
        self._TakeoffAlt = [10.0]
        """
        The minimum relative altitude the model will move to before RTL. Set to
        zero to return at current altitude.
         - Name: RTLAlt RTLAlt
         - DefaultValue: 1500
         - Unit: cm
         - Range: 0<=x<=8000
        """
        self._RTLAlt = [1500]
        """
        This is the altitude the vehicle will move to as the final stage of RTL or
        after completing a mission. Set to zero to land.
         - Name: RTLAltFinal RTLAltFinal
         - DefaultValue: 0
         - Unit: cm
         - Range: -1<=x<=1000
        """
        self._RTLAltFinal = [0]
        """
        Waypoint horizontal speed target.
        (This parameter is an alternative to "RTL_SPEED".)
         - Name: WpNavSpeed WpNavSpeed
         - DefaultValue: 500
         - Unit: cm/s
         - Range: 0<=x<=2000
        """
        self._WpNavSpeed = [500]
        """
        Time  to loiter above home before beginning final descent.
         - Name: RTLLoitTime RTLLoitTime
         - DefaultValue: 5000
         - Unit: msec
         - Range: 0<=x<=60000
        """
        self._RTLLoitTime = [5000]
        """
        Waypoint descent speed target.
        (This parameter is an alternative to "LAND_SPEED_HIGH".)
         - Name: WpNavSpeedDn WpNavSpeedDn
         - DefaultValue: 100
         - Unit: cm/s
         - Range: 0<=x<=500
        """
        self._WpNavSpeedDn = [100]
        """
        The descent speed for the final stage of landing in cm/s.
         - Name: LandSpeed LandSpeed
         - DefaultValue: 50
         - Unit: cm/s
         - Range: 30<=x<=200
        """
        self._LandSpeed = [50]
        """
        Delay before automatic disarm in seconds. A value of zero disables auto
        disarm.
         - Name: DisarmDelay DisarmDelay
         - DefaultValue: 30
         - Unit: sec
         - Range: 0<=x<=127
        """
        self._DisarmDelay = [30]
        """
        This RTC has different behaviors depending on the system, so select which
        system to behave.
         - Name:  SystemBehavior
         - DefaultValue: single
        """
        self._SystemBehavior = ['single']
        
        # </rtc-template>

        self._vehicle = None  # vehicle object
        self._flight_time_start = 0  # Flight Start Time
        self._drone_sts = STS_DISARM
        self._comm_err_timer = 0  # communication error check timer
        self._takeoff_alt = 0
        self._tgt_alt = 0
        self._rtl_cause = 0
        self._rtl_stage = 0
         
    ##
    # RTC Initialize
    #
    # The initialize action (on CREATED->ALIVE transition)
    # formaer rtc_init_entry() 
    # 
    # @return RTC::ReturnCode_t
    # 
    #
    def onInitialize(self):
        # Bind variables and configuration variable
        self.bindParameter("BattCell", self._BattCell, "4")
        self.bindParameter("RTLLowBatt", self._RTLLowBatt, "30")
        self.bindParameter("BattChk", self._BattChk, "on")
        self.bindParameter("TakeoffAlt", self._TakeoffAlt, "10.0")
        self.bindParameter("RTLAlt", self._RTLAlt, "1500")
        self.bindParameter("RTLAltFinal", self._RTLAltFinal, "0")
        self.bindParameter("WpNavSpeed", self._WpNavSpeed, "500")
        self.bindParameter("RTLLoitTime", self._RTLLoitTime, "5000")
        self.bindParameter("WpNavSpeedDn", self._WpNavSpeedDn, "100")
        self.bindParameter("LandSpeed", self._LandSpeed, "50")
        self.bindParameter("DisarmDelay", self._DisarmDelay, "30")
        self.bindParameter("SystemBehavior", self._SystemBehavior, "single")
        
        # Set InPort buffers
        self.addInPort("inTargetPos",self._inTargetPosIn)
        self.addInPort("inTargetSpeed",self._inTargetSpeedIn)
        self.addInPort("inTargetYaw",self._inTargetYawIn)
        
        # Set OutPort buffers
        self.addOutPort("outGPSHome",self._outGPSHomeOut)
        self.addOutPort("outGPSDrone",self._outGPSDroneOut)
        self.addOutPort("outBattV",self._outBattVOut)
        self.addOutPort("outBattA",self._outBattAOut)
        self.addOutPort("outFltTime",self._outFltTimeOut)
        self.addOutPort("outGyro",self._outGyroOut)
        self.addOutPort("outAccel",self._outAccelOut)
        self.addOutPort("outMag",self._outMagOut)
        self.addOutPort("outTemp",self._outTempOut)
        self.addOutPort("outPres",self._outPresOut)
        self.addOutPort("outFltMode",self._outFltModeOut)
        self.addOutPort("outState",self._outStateOut)
        self.addOutPort("outSysErr",self._outSysErrOut)
        self.addOutPort("outRawIMU",self._outRawIMUOut)
        self.addOutPort("outCorrectIMU",self._outCorrectIMUOut)
        
        # Set service provider to Ports
        
        # Set service consumers to Ports
        
        # Set CORBA Service Ports
        
        return RTC.RTC_OK
    
        ##
        # Close vehicle object
        # 
        # The finalize action (on ALIVE->END transition)
        # formaer rtc_exiting_entry()
        # 
        # @return RTC::ReturnCode_t
    
        # 
    def onFinalize(self):
    
        self._vehicle.close()
        logger.info("vehicle close")

        return RTC.RTC_OK
    
        ##
        # - Connecting to the vehicle(default target adress: 127.0.0.1:14550).
    #    - Download the "home_location"
        #
        # The startup action when ExecutionContext startup
        # former rtc_starting_entry()
        # 
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onStartup(self, ec_id):
    
        rtn = False

        if self._is_vehicle_connected() is True:
            if self._is_home_location_acquired() is True:
                rtn = True

        if rtn is True:
            return RTC.RTC_OK
        else:
            return RTC.RTC_ERROR
    
    #    ##
    #    #
    #    # The shutdown action when ExecutionContext stop
    #    # former rtc_stopping_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    #def onShutdown(self, ec_id):
    #
    #    return RTC.RTC_OK
    
        ##
        # - Switch to armed and wait for reflection.
    #    - Vehicle mode is set to GUIDED
    #    - Parameter setting
    #    - Start the counting of the flight timer.
        #
        # The activated action (Active state entry action)
        # former rtc_active_entry()
        #
        # @param ec_id target ExecutionContext Id
        # 
        # @return RTC::ReturnCode_t
        #
        # @pre "home_location" exists
        #
    def onActivated(self, ec_id):
    
        rtn = False

        # low-battery threshold calculation
        self._batt_thr = (BATT_LIPO_VPC - (BATT_LIPO_VPC_PERCENT -
                                           (self._RTLLowBatt[0] / 100.0))) * self._BattCell[0]

        # battery check
        if self._is_batt_state() is True:
            # RTL or not
            if self._drone_sts != STS_RTL:
                # home_location check
                if self._vehicle.home_location is not None:
                    # Change the "GUIDED" mode
                    self._vehicle.mode = VehicleMode("GUIDED")

                    if self._switching_to_armed() is True:
                        # Battery parameters tuning
                        self._batt_param_tuning()
                        # Takeoff paramter tuning
                        self._vehicle.parameters[
                            'DISARM_DELAY'] = self._DisarmDelay[0]
                        # RTL paramter tuning
                        self._rtl_param_tuning()
                        # Change the arm status
                        self._drone_sts = STS_ARMED
                        self._write_io_state()
                        # Error Info clear
                        self._write_io_err_inf("")
                        self._init_val_rtl()
                        # Flight time count to start
                        self._flight_time_start = time.time()
                        rtn = True
                    else:
                        self._write_io_err_inf(ERR_MSG_005)
                else:
                    self._write_io_err_inf(ERR_MSG_004)
            else:
                # in RTL
                rtn = True

        if rtn is True:
            return RTC.RTC_OK
        else:
            return RTC.RTC_ERROR
    
        ##
        # RTL if in flight, otherwise switch to disarm if RTL is done.
        #
        # The deactivated action (Active state exit action)
        # former rtc_active_exit()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onDeactivated(self, ec_id):
    
        if self._drone_sts != STS_RTL:
            if self._drone_sts not in [STS_DISARM, STS_ARMED]:
                self._rtl_by_err()
                self._write_flt_mode()
                logger.info("RTC deactivated. Start the RTL.")
            else:
                self._vehicle.armed = False
                self._drone_sts = STS_DISARM
                logger.info("RTC deactivated. "
                            "Switching to disarm.")
            self._write_io_state()
        else:
            if self._vehicle.armed is True:
                if self._SystemBehavior[0] == 'formation':
                    if self._vehicle.mode.name != "RTL":
                        self._rtl_by_err()
                        self._write_flt_mode()
                        logger.info("RTC deactivated. Start the RTL.")
                    else:
                        logger.info("RTC deactivated. "
                                    "Will be continued RTL.")
                else:
                    logger.info("RTC deactivated. "
                                "Will be continued RTL.")
            else:
                self._drone_sts = STS_DISARM
                self._write_io_state()
                logger.info("Landed in the home position.")

        return RTC.RTC_OK
    
        ##
        # - Get and output of the sensor info through the DroneKit-Python API.
    #    - Flight instruction through the DronKit-Python API.
        #
        # The execution action that is invoked periodically
        # former rtc_active_do()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onExecute(self, ec_id):
    
        rtn = True

        # Armed Check
        if self._vehicle.armed is True:
            # RTL or not
            if self._drone_sts != STS_RTL:
                # battery check
                if self._is_batt_state() is True:
                    # Read the target position
                    self._read_target_pos()
                    self._read_target_spd()
                    self._read_target_yaw()
                    if (self._target_pos and
                            self._target_spd and
                            self._target_yaw):
                        self._comm_err_timer = 0
                        self._flight_ctrl_core()
                    else:
                        # Communication error check
                        self._comm_err_timer += 1
                        if self._comm_err_timer == COMM_ERR_TIMER:
                            # Start the RTL
                            self._comm_err_timer = 0
                            self._write_io_err_inf(ERR_MSG_006)
                            self._rtl_by_err()

                    # judge the target position arrive
                    self._arv_judge()
                else:
                    # Start the RTL
                    self._rtl_by_err()
            else:
                if self._SystemBehavior[0] == 'formation':
                    if self._rtl_cause == RTL_CAUSE_BY_MANU:
                        self._continuous_manu_rtl()

            # Read & output the sensor info
            self._write_info()
        else:
            # switching from armed to disarm
            self.deactivate(ec_id)
            logger.info("onExecute active->deactivate")

        if rtn is True:
            return RTC.RTC_OK
        else:
            return RTC.RTC_ERROR
    
    #    ##
    #    #
    #    # The aborting action when main logic error occurred.
    #    # former rtc_aborting_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    #def onAborting(self, ec_id):
    #
    #    return RTC.RTC_OK
    
        ##
        # RTL if in flight
        #
        # The error action in ERROR state
        # former rtc_error_do()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onError(self, ec_id):
    
        if self._drone_sts not in [STS_DISARM, STS_ARMED, STS_RTL]:
            self._rtl_by_err()

        return RTC.RTC_OK
    
    #    ##
    #    #
    #    # The reset action that is invoked resetting
    #    # This is same but different the former rtc_init_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    #def onReset(self, ec_id):
    #
    #    return RTC.RTC_OK
    
    #    ##
    #    #
    #    # The state update action that is invoked after onExecute() action
    #    # no corresponding operation exists in OpenRTm-aist-0.2.0
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #

    #    #
    #def onStateUpdate(self, ec_id):
    #
    #    return RTC.RTC_OK
    
    #    ##
    #    #
    #    # The action that is invoked when execution context's rate is changed
    #    # no corresponding operation exists in OpenRTm-aist-0.2.0
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    #def onRateChanged(self, ec_id):
    #
    #    return RTC.RTC_OK
    
    def _is_vehicle_connected(self):

        rtn = False

        try:
            logger.info("\nConnecting to vehicle on: %s", TARGET_IP)
            self._vehicle = connect(
                TARGET_IP, wait_ready=True, vehicle_class=MyVehicle)
            rtn = True
        except socket.error:
            logger.error(ERR_MSG_001)
        except dronekit.APIException:
            logger.error(ERR_MSG_002)

        return rtn

    def _is_home_location_acquired(self):

        rtn = True

        try:
            retry_cnt = 0
            while retry_cnt < HOME_LOC_GET_RETRY_CNT:
                cmds = self._vehicle.commands
                cmds.download()
                cmds.wait_ready()
                if not self._vehicle.home_location:
                    retry_cnt += 1
                    logger.info("Waiting for home location...")
                    time.sleep(1)
                else:
                    logger.info("%s", self._vehicle.home_location)
                    break

            if retry_cnt == HOME_LOC_GET_RETRY_CNT:
                if not self._vehicle.home_location:
                    logger.error(ERR_MSG_003)
                    rtn = False

        except:
            logger.error("[E]Please re-startup this RTC.")
            rtn = False

        return rtn

    def _switching_to_armed(self):

        rtn = True

        self._vehicle.armed = True
        retrycnt = 0
        while retrycnt < ARMED_RETRY_CNT:
            if self._vehicle.armed is not True:
                retrycnt += 1
                logger.info("Waiting for arming...")
                time.sleep(1)
            else:
                logger.info("Switching to armed success.")
                break

        if retrycnt == ARMED_RETRY_CNT:
            if self._vehicle.armed is not True:
                rtn = False

        return rtn

    def _batt_param_tuning(self):

        self._vehicle.parameters['BATT_MONITOR'] = BATT_MONI_V_ONLY
        self._vehicle.parameters['BATT_VOLT_PIN'] = BATT_V_PIN_PIXHAWK
        self._vehicle.parameters['BATT_CURR_PIN'] = BATT_C_PIN_PIXHAWK

    def _rtl_param_tuning(self):

        self._vehicle.parameters['RTL_ALT'] = self._RTLAlt[0]
        self._vehicle.parameters['RTL_ALT_FINAL'] = self._RTLAltFinal[0]
        self._vehicle.parameters['WPNAV_SPEED'] = self._WpNavSpeed[0]
        self._vehicle.parameters['WPNAV_SPEED_DN'] = self._WpNavSpeedDn[0]
        self._vehicle.parameters['RTL_LOIT_TIME'] = self._RTLLoitTime[0]
        self._vehicle.parameters['LAND_SPEED'] = self._LandSpeed[0]

    def _get_propo_fltmode(self):

        fltmd_val = COPTER_PARAM_FLTMODE_GUIDED

        if 0 <= self._vehicle.channels['5'] <= 1230:
            fltmd_val = self._vehicle.parameters['FLTMODE1']
        elif 1231 <= self._vehicle.channels['5'] <= 1749:
            fltmd_val = self._vehicle.parameters['FLTMODE2']
        elif 1750 <= self._vehicle.channels['5']:
            fltmd_val = self._vehicle.parameters['FLTMODE6']
        else:
            pass

        return fltmd_val

    def _read_target_pos(self):

        self._target_pos = False

        if self._inTargetPosIn.isNew():
            self._d_inTargetPos = self._inTargetPosIn.read()
            self._target_pos = True

    def _read_target_spd(self):

        self._target_spd = False

        if self._inTargetSpeedIn.isNew():
            self._d_inTargetSpeed = self._inTargetSpeedIn.read()
            self._target_spd = True

    def _read_target_yaw(self):

        self._target_yaw = False

        if self._inTargetYawIn.isNew():
            self._d_inTargetYaw = self._inTargetYawIn.read()
            self._target_yaw = True

    def _flight_ctrl_core(self):

        # Setting the air speed
        if self._d_inTargetSpeed.data != 0:
            if self._vehicle.airspeed != self._d_inTargetSpeed.data:
                self._vehicle.airspeed = self._d_inTargetSpeed.data
        else:
            self._vehicle.airspeed = VEHICLE_DEF_AIR_SPD

        if self._drone_sts == STS_ARMED:
            if self._d_inTargetPos.altitude == self._TakeoffAlt[0]:
                # take off
                self._takeoff_alt = self._d_inTargetPos.altitude
                self._vehicle.simple_takeoff(self._takeoff_alt)
                self._drone_sts = STS_TAKEOFF
                logger.info("Start the takeoff")
            else:
                logger.error("[E]Wrong takeoff parameter: %f[m]",
                             self._d_inTargetPos.altitude)
        elif self._drone_sts in [STS_IN_MOVE, STS_ALT_CHG, STS_HOVERING, STS_YAW_CHG]:
            if self._is_rtl_req() is True:
                # Go to RTLAlt
                logger.info("System Behavior is %s", self._SystemBehavior[0])
                if self._SystemBehavior[0] == 'formation':
                    self._drone_sts = STS_RTL
                    self._rtl_cause = RTL_CAUSE_BY_MANU
                    self._rtl_stage = RTL_STAGE_1ST
                    self._tgt_alt = self._RTLAlt[0] / 100
                    point = LocationGlobalRelative(
                        self._d_outGPSDrone.latitude,
                        self._d_outGPSDrone.longitude,
                        self._tgt_alt)
                    self._vehicle.simple_goto(point)
                    logger.info("Start the RTL by manually.")
                else:
                    self._start_rtl()
            else:
                self._move_to_target_pos()
                if self._d_inTargetYaw.data[0] != 0:
                    self._condition_yaw()
        else:
            # None, if the state is RTL or take off
            pass

    def _is_rtl_req(self):

        rtn = False

        if self._d_inTargetPos.latitude == self._d_outGPSHome.latitude:
            if self._d_inTargetPos.longitude == self._d_outGPSHome.longitude:
                if self._d_inTargetPos.altitude == self._d_outGPSHome.altitude:
                    rtn = True

        return rtn

    def _move_to_target_pos(self):
        logger.debug("Drone lat:%.10f lon:%.10f alt:%.10f",
                     self._d_outGPSDrone.latitude,
                     self._d_outGPSDrone.longitude,
                     self._d_outGPSDrone.altitude)

        logger.debug("Target lat:%.10f lon:%.10f alt:%.10f",
                     self._d_inTargetPos.latitude,
                     self._d_inTargetPos.longitude,
                     self._d_inTargetPos.altitude)

        dist_cal = drone_distance_cal.DroneMathmetics()
        dist = dist_cal.distance_math(self._d_outGPSDrone.longitude,
                                      self._d_inTargetPos.longitude,
                                      self._d_outGPSDrone.latitude,
                                      self._d_inTargetPos.latitude,
                                      0,
                                      0)

        logger.debug("Target dist:%f[m]", dist)

        # "MIN_MV_DIST" means mininum move distance
        # See "機能仕様書_DroneController" for details
        if dist <= MIN_MV_DIST:
            # altitude change, if change value is allowable range
            self._alt_chg_req_alt = self._d_inTargetPos.altitude
            if abs(self._d_inTargetPos.altitude - self._d_outGPSDrone.altitude) > 0:
                self._drone_sts = STS_ALT_CHG
                logger.debug("alt change: %f[m]", self._d_inTargetPos.altitude)
            else:
                self._drone_sts = STS_HOVERING
        else:
            self._move_target_lat = self._d_inTargetPos.latitude
            self._move_target_lon = self._d_inTargetPos.longitude
            self._drone_sts = STS_IN_MOVE

        # Move to the target point
        point = LocationGlobalRelative(
            self._d_inTargetPos.latitude,
            self._d_inTargetPos.longitude,
            self._d_inTargetPos.altitude)
        self._vehicle.simple_goto(point)

    def _condition_yaw(self):

        # create the CONDITION_YAW command using command_long_encode()
        msg = self._vehicle.message_factory.command_long_encode(
            0, 0,  # target system, target component
            mavutil.mavlink.MAV_CMD_CONDITION_YAW,  # command
            0,  # confirmation
            self._d_inTargetYaw.data[1],  # param 1, yaw in degrees
            0,                            # param 2, yaw speed deg/s
            self._d_inTargetYaw.data[0],  # param 3, direction -1 ccw, 1 cw
            0,                            # param 4, abusolute direction
            0, 0, 0)                      # param 5 ~ 7 not used

        # send command to vehicle
        self._vehicle.send_mavlink(msg)

        self._drone_sts = STS_YAW_CHG

    def _arv_judge(self):

        if self._drone_sts == STS_TAKEOFF:
            if self._takeoff_alt * TAKEOFF_ARV_COEF <= self._d_outGPSDrone.altitude:
                logger.info("takeoff complete")
                self._drone_sts = STS_HOVERING

    def _init_val_rtl(self):

        self._rtl_cause = 0
        self._rtl_stage = 0

    def _start_rtl(self):

        self._vehicle.mode = VehicleMode("RTL")
        self._drone_sts = STS_RTL

    def _continuous_manu_rtl(self):

        if self._rtl_stage == RTL_STAGE_1ST:
            if self._is_arv_tgt_alt() is True:
                self._rtl_stage = RTL_STAGE_2ND
                point = LocationGlobalRelative(
                    self._d_outGPSHome.latitude,
                    self._d_outGPSHome.longitude,
                    self._tgt_alt)
                self._vehicle.simple_goto(point)

        elif self._rtl_stage == RTL_STAGE_2ND:
            self._read_target_pos()
            if self._target_pos:
                if self._is_rtl_req():
                    self._rtl_stage = RTL_STAGE_3RD
                    self._start_rtl()

        else:
            pass

    def _rtl_by_err(self):

        self._rtl_cause = RTL_CAUSE_BY_ERR
        self._start_rtl()

    def _is_arv_tgt_alt(self):

        rtn = False

        tgt_pos_min = self._tgt_alt * TGT_ALT_ARV_COEF_MIN
        tgt_pos_max = self._tgt_alt * TGT_ALT_ARV_COEF_MAX
        if tgt_pos_min <= self._d_outGPSDrone.altitude <= tgt_pos_max:
            logger.info("Arrived at the target alt")
            rtn = True

        return rtn

    def _write_info(self):

        self._write_gps_home()
        self._write_gps_drone()
        self._write_batt()
        self._write_flt_time()
        self._write_gyro()
        self._write_accel()
        self._write_mag()
        self._write_temperature()
        self._write_pressure()
        self._write_flt_mode()
        self._write_io_state()
        self._write_raw_imu()
        self._write_correct_imu()

    def _write_gps_home(self):

        if self._vehicle.home_location is not None:
            tmp_home_location = self._vehicle.home_location
            self._d_outGPSHome.tm.sec = 0
            self._d_outGPSHome.tm.nsec = 0
            self._d_outGPSHome.timeFromGPS.sec = 0
            self._d_outGPSHome.timeFromGPS.msec = 0
            self._d_outGPSHome.latitude = tmp_home_location.lat
            self._d_outGPSHome.longitude = tmp_home_location.lon
            self._d_outGPSHome.altitude = tmp_home_location.alt
            self._d_outGPSHome.horizontalError = 0.0
            self._d_outGPSHome.verticalError = 0.0
            self._d_outGPSHome.heading = 0.0
            self._d_outGPSHome.horizontalSpeed = 0.0
            self._d_outGPSHome.verticalSpeed = 0.0
            self._d_outGPSHome.numSatellites = 0
            self._d_outGPSHome.fixType = RTC.GPS_FIX_NORMAL
            self._outGPSHomeOut.write()

    def _write_gps_drone(self):

        if self._vehicle.location.global_relative_frame.lat is not None:
            if self._vehicle.location.global_relative_frame.lon is not None:
                if self._vehicle.location.global_relative_frame.alt is not None:
                    self._d_outGPSDrone.latitude = float(
                        self._vehicle.location.global_relative_frame.lat)
                    self._d_outGPSDrone.longitude = float(
                        self._vehicle.location.global_relative_frame.lon)
                    self._d_outGPSDrone.altitude = float(
                        self._vehicle.location.global_relative_frame.alt)
                    self._d_outGPSDrone.tm.sec = 0
                    self._d_outGPSDrone.tm.nsec = 0
                    self._d_outGPSDrone.timeFromGPS.sec = 0
                    self._d_outGPSDrone.timeFromGPS.msec = 0
                    self._d_outGPSDrone.horizontalError = 0.0
                    self._d_outGPSDrone.verticalError = 0.0
                    self._d_outGPSDrone.heading = 0.0
                    self._d_outGPSDrone.horizontalSpeed = 0.0
                    self._d_outGPSDrone.verticalSpeed = 0.0
                    self._d_outGPSDrone.numSatellites = 0
                    self._d_outGPSDrone.fixType = RTC.GPS_FIX_NORMAL
                    self._outGPSDroneOut.write()
                    logger.debug(
                        "%s", self._vehicle.location.global_relative_frame)

    def _write_batt(self):
        # Voltage
        if self._vehicle.battery.voltage is not None:
            self._d_outBattV.data = self._vehicle.battery.voltage
            self._outBattVOut.write()

        # Current
        if self._vehicle.battery.current is not None:
            self._d_outBattA.data = self._vehicle.battery.current
            self._outBattAOut.write()

        logger.debug("Batt: %s V, %s mA",
                     self._d_outBattV.data, self._d_outBattA.data)

    def _write_flt_time(self):

        current_flight_time = time.time() - self._flight_time_start
        self._d_outFltTime.sec = int(current_flight_time)
        self._outFltTimeOut.write()

        min = self._d_outFltTime.sec / 60
        sec = self._d_outFltTime.sec % 60
        logger.debug("Time: %0d:%02d", min, sec)

    def _write_gyro(self):

        if (self._vehicle.attitude.roll is not None and
                self._vehicle.attitude.pitch is not None and
                self._vehicle.attitude.yaw is not None):
            self._d_outGyro.data.r = self._vehicle.attitude.roll
            self._d_outGyro.data.p = self._vehicle.attitude.pitch
            self._d_outGyro.data.y = self._vehicle.attitude.yaw
            self._outGyroOut.write()
            logger.debug("%s", self._vehicle.attitude)

    def _write_accel(self):

        accel = self._vehicle.velocity
        if all(elem is not None for elem in accel):
            self._d_outAccel.data = accel
            self._outAccelOut.write()
            logger.debug("Accel: %s", self._vehicle.velocity)

    def _write_mag(self):

        if self._vehicle.heading is not None:
            self._d_outMag.data = self._vehicle.heading
            self._outMagOut.write()
            logger.debug("Mag:%d", self._d_outMag.data)

    def _write_temperature(self):

        temp = self._vehicle.scaled_pres.temperature
        if temp is not None:
            self._d_outTemp.data = temp
            self._outTempOut.write()
            logger.debug("Temperature:%s[C]", self._d_outTemp.data)

    def _write_pressure(self):

        pres = self._vehicle.scaled_pres.press_abs
        if pres is not None:
            self._d_outPres.data = pres
            self._outPresOut.write()
            logger.debug("Pressure:%s[hPa]", self._d_outPres.data)

    def _write_flt_mode(self):

        self._d_outFltMode.data = self._vehicle.mode.name
        self._outFltModeOut.write()
        logger.debug("FltMode:%s", self._d_outFltMode.data)

    def _write_io_state(self):

        self._d_outState.data = droneio_sts_str[self._drone_sts]
        self._outStateOut.write()
        logger.debug("State:%s", self._d_outState.data)

    def _write_io_err_inf(self, msg):

        self._d_outSysErr.data = msg
        self._outSysErrOut.write()
        logger.error("ErrMsg:%s", msg)

    def _write_raw_imu(self):

        self._d_outRawIMU.data = [self._vehicle.raw_imu.xacc,
                                  self._vehicle.raw_imu.yacc,
                                  self._vehicle.raw_imu.zacc,
                                  self._vehicle.raw_imu.xgyro,
                                  self._vehicle.raw_imu.ygyro,
                                  self._vehicle.raw_imu.zgyro,
                                  self._vehicle.raw_imu.xmag,
                                  self._vehicle.raw_imu.ymag,
                                  self._vehicle.raw_imu.zmag]
        self._outRawIMUOut.write()
        logger.debug("RawIMU:%s", self._d_outRawIMU.data)

    def _write_correct_imu(self):

        accel = self._vehicle.velocity
        if accel[0] is None:
            accel[0] = 0.0

        if accel[1] is None:
            accel[1] = 0.0

        if accel[2] is None:
            accel[2] = 0.0

        if self._vehicle.attitude.roll is not None:
            roll = self._vehicle.attitude.roll
        else:
            roll = 0.0

        if self._vehicle.attitude.pitch is not None:
            pitch = self._vehicle.attitude.pitch
        else:
            pitch = 0.0

        if self._vehicle.attitude.yaw is not None:
            yaw = self._vehicle.attitude.yaw
        else:
            yaw = 0.0

        self._d_outCorrectIMU.data = [
            accel[0], accel[1], accel[2],
            roll, pitch, yaw,
            self._vehicle.heading]
        self._outCorrectIMUOut.write()
        logger.debug("CorrectIMU:%s", self._d_outCorrectIMU.data)

    def _is_batt_state(self):

        rtn = True

        if self._BattChk[0] == 'on':
            if(self._vehicle.battery.voltage < self._batt_thr):
                self._write_io_err_inf(ERR_MSG_007)
                rtn = False

        return rtn


class RawIMU(object):

    def __init__(self, time_boot_us=0.0, xacc=0.0, yacc=0.0, zacc=0.0, xgyro=0.0, ygyro=0.0, zgyro=0.0, xmag=0.0, ymag=0.0, zmag=0.0):
        self._time_boot_us = time_boot_us
        self._xacc = xacc
        self._yacc = yacc
        self._zacc = zacc
        self._xgyro = xgyro
        self._ygyro = ygyro
        self._zgyro = zgyro
        self._xmag = xmag
        self._ymag = ymag
        self._zmag = zmag

    def __str__(self):
        return "xacc={},yacc={},zacc={},xgyro={},ygyro={},zgyro={},xmag={},ymag={},zmag={}".format(self._xacc, self._yacc, self._zacc, self._xgyro, self._ygyro, self._zgyro, self._xmag, self._ymag, self._zmag)


class Scaled_Pressure(object):

    def __init__(self, time_boot_ms=0.0, press_abs=0.0, press_diff=0.0, temperature=0.0):
        self._time_boot_ms = time_boot_ms
        self._press_abs = press_abs
        self._press_diff = press_diff
        self._temperature = temperature

    def __str__(self):
        return "press_abs={},temperature={}".format(self._press_abs, self._temperature)


class MyVehicle(Vehicle):

    def __init__(self, *args):
        super(MyVehicle, self).__init__(*args)

        # Create an Vehicle.raw_imu object with initial values set to None.
        self._raw_imu = RawIMU()

        # Create a message listener using the decorator.
        @self.on_message('RAW_IMU')
        def listener(self, name, message):
            self._raw_imu.time_boot_us = message.time_usec
            self._raw_imu.xacc = message.xacc
            self._raw_imu.yacc = message.yacc
            self._raw_imu.zacc = message.zacc
            self._raw_imu.xgyro = message.xgyro
            self._raw_imu.ygyro = message.ygyro
            self._raw_imu.zgyro = message.zgyro
            self._raw_imu.xmag = message.xmag
            self._raw_imu.ymag = message.ymag
            self._raw_imu.zmag = message.zmag

            self.notify_attribute_listeners('raw_imu', self._raw_imu)

        # Create an Vehicle.'scaled_pres' object with initial values set to
        # None.
        self._scaled_pres = Scaled_Pressure()

        # Create a message listener using the decorator.
        @self.on_message('SCALED_PRESSURE')
        def listener(self, name, message):
            self._scaled_pres.time_boot_ms = message.time_boot_ms
            self._scaled_pres.press_abs = message.press_abs
            self._scaled_pres.press_diff = message.press_diff
            self._scaled_pres.temperature = message.temperature * 0.01

            self.notify_attribute_listeners('scaled_pres', self._scaled_pres)

    @property
    def raw_imu(self):
        return self._raw_imu

    @property
    def scaled_pres(self):
        return self._scaled_pres


def DroneIOInit(manager):
    profile = OpenRTM_aist.Properties(defaults_str=droneio_spec)
    manager.registerFactory(profile,
                            DroneIO,
                            OpenRTM_aist.Delete)

def MyModuleInit(manager):
    DroneIOInit(manager)

    # Create a component
    comp = manager.createComponent("DroneIO")

def main():
    mgr = OpenRTM_aist.Manager.init(sys.argv)
    mgr.setModuleInitProc(MyModuleInit)
    mgr.activateManager()
    mgr.runManager()

if __name__ == "__main__":
    main()

